#include "CoherentHierarchicalCullingManager.h"
#include "CullingLogManager.h"

#include <time.h>

namespace GtpVisibility {

//-----------------------------------------------------------------------
CoherentHierarchicalCullingManager::CoherentHierarchicalCullingManager()
{
	SetAssumedVisibility(0);
	// initialise random generator in case we use assumed visibility
	//srand(time(NULL));
}
//-----------------------------------------------------------------------
CoherentHierarchicalCullingManager::CoherentHierarchicalCullingManager(
														const unsigned int assumedVisibility)
{
	SetAssumedVisibility(assumedVisibility);
	// initialise random generator in case we use assumed visibility
	//srand(time(NULL));
}
//-----------------------------------------------------------------------
void CoherentHierarchicalCullingManager::RenderScene()
{
	if (0) CullingLogManager::GetSingleton()->LogMessage("chc");
	QueryQueue queryQueue;
	unsigned int visiblePixels = 0;
	
	/////////////
	//-- PART 1: process finished occlusion queries

	while (!mHierarchyInterface->GetQueue()->empty() || !queryQueue.empty())
	{
		//////////
		//-- only wait for result if there are no nodes to process
		while (!queryQueue.empty() && 
			   (NodeInvalid(queryQueue.front().first) ||
			    queryQueue.front().second->GetQueryResult(
				             visiblePixels, mHierarchyInterface->GetQueue()->empty())))
		{
			HierarchyNode *node = queryQueue.front().first;
			queryQueue.pop();

			// parent was tested invisible => remove children from queue
			if (NodeInvalid(node))
			{
				//CullingLogManager::GetSingleton()->LogMessage("skipping this node");
				continue;
			}
			// tested visible
			if (visiblePixels > mVisibilityThreshold)
			{
				// in case geometry is in interior node: ensure that we only traverse once
				if (!mHierarchyInterface->IsNodeVisible(node))
				{
					mHierarchyInterface->TraverseNode(node);
				}

				mHierarchyInterface->PullUpVisibility(node);
			}
			else
			{	
				mHierarchyInterface->SetNodeVisible(node, false);

                ++ mNumQueryCulledNodes;
				
				if (mVisualizeCulledNodes)
				{
					mHierarchyInterface->VisualizeCulledNode(node, QUERY_CULLED);
				}
			}
			
			// update node's visited flag
			mHierarchyInterface->SetLastVisited(node, mHierarchyInterface->GetFrameId());	
		}
		
		//-- PART 2: hierarchical traversal
		if (!mHierarchyInterface->GetQueue()->empty())
		{
			HierarchyNode *node = mHierarchyInterface->GetQueue()->top();
			mHierarchyInterface->GetQueue()->pop();
				
			// parent was tested invisible => remove children from queue
			if (NodeInvalid(node))
			{
				//CullingLogManager::GetSingleton()->LogMessage("skipping this node");
				continue;
			}

			bool intersects = false;

			if (!mHierarchyInterface->CheckFrustumVisible(node, intersects))
			{
				++ mNumFrustumCulledNodes;

				if (mVisualizeCulledNodes)
				{
					mHierarchyInterface->VisualizeCulledNode(node, FRUSTUM_CULLED);
				}
			}
			//-- if node intersects near plane, skip query because wrong results possible
			else if (intersects)
			{
				SkipQuery(node);
			}
			else
			{
				// identify previously visible nodes
				bool wasVisible = mHierarchyInterface->IsNodeVisible(node) && 
					(mHierarchyInterface->LastVisited(node) == mHierarchyInterface->GetFrameId() - 1);
				
				// if we assume node to be visible in this frame => skip query 
				const bool skipQuery = wasVisible && 
					                   (mAssumedVisibility > 0) && 
									   DecideVisible(node) && 
									   mHierarchyInterface->HasGeometry(node);

				if (skipQuery)
				{
					SkipQuery(node);
					continue;
				}

                // identify nodes that we cannot skip queries for
				// geometry not only in leaves => test for renderable geometry
				bool issueQuery = !wasVisible || mHierarchyInterface->HasGeometry(node);
							
				// reset node's visibility classification
				// set visible if geometry in node so we only traverse once
				mHierarchyInterface->SetNodeVisible(node, wasVisible && issueQuery);

				// skip testing previously visible nodes without geometry
				if (issueQuery)
				{
					++ mNumQueriesIssued;
					
					queryQueue.push(QueryPair(node, mHierarchyInterface->
						IssueNodeOcclusionQuery(node, wasVisible)));
				}
				else
				{
					// update node's visited flag
					mHierarchyInterface->SetLastVisited(node, mHierarchyInterface->GetFrameId());
				}

				// always traverse a node if it was visible
				if (wasVisible)
				{
					mHierarchyInterface->TraverseNode(node);
				}
			}
		}
	}
}
//-----------------------------------------------------------------------
void CoherentHierarchicalCullingManager::SetAssumedVisibility(const unsigned int assumedVisibility)
{
	mAssumedVisibility = assumedVisibility;
	
	if (!mAssumedVisibility)
	{
		mThreshold = 0;
	}
	else
	{
		mThreshold = RAND_MAX - RAND_MAX / mAssumedVisibility;
		 // fix visibility
		if (mAssumedVisibility > 100)
			mThreshold = RAND_MAX;
	}
	
}
//-----------------------------------------------------------------------
inline bool CoherentHierarchicalCullingManager::DecideVisible(HierarchyNode *node) const
{
	return rand() < mThreshold;
}
//-----------------------------------------------------------------------
inline void CoherentHierarchicalCullingManager::SkipQuery(HierarchyNode *node) const
{
	//-- set node to be visible in this frame, then traverse it
	mHierarchyInterface->SetLastVisited(node, mHierarchyInterface->GetFrameId());
	
	mHierarchyInterface->PullUpVisibility(node);			
	mHierarchyInterface->TraverseNode(node);
}
//-----------------------------------------------------------------------
bool CoherentHierarchicalCullingManager::NodeInvalid(HierarchyNode *node) const
{
	// parent was tested invisible in this frame
	HierarchyNode *parent = mHierarchyInterface->GetParent(node);
    return 
		parent && 
		(mHierarchyInterface->LastVisited(node) == mHierarchyInterface->GetFrameId()) && 
		!mHierarchyInterface->IsNodeVisible(parent);
}
} // namespace GtpVisibility
